# -*- coding: utf-8 -*-

"""
Serialization of PENMAN graphs.
"""

from typing import Union, Iterable, Iterator, List, IO
from pathlib import Path

from penman.types import (
    Variable,
    BasicTriple,
)
from penman.tree import Tree
from penman.graph import Graph
from penman.model import Model
from penman._parse import (
    parse,
    iterparse,
    parse_triples,
)
from penman._format import (
    format,
    format_triples,
)
from penman import layout


# "Utility" types; not Penman-specific

FileOrFilename = Union[str, Path, IO[str]]


class PENMANCodec(object):
    """
    An encoder/decoder for PENMAN-serialized graphs.
    """

    def __init__(self, model: Model = None):
        if model is None:
            model = Model()
        self.model = model

    def decode(self, s: str) -> Graph:
        """
        Deserialize PENMAN-notation string *s* into its Graph object.

        Args:
            s: a string containing a single PENMAN-serialized graph
        Returns:
            The :class:`~penman.graph.Graph` object described by *s*.
        Example:
            >>> from penman.codec import PENMANCodec
            >>> codec = PENMANCodec()
            >>> codec.decode('(b / bark-01 :ARG0 (d / dog))')
            <Graph object (top=b) at ...>
        """
        tree = parse(s)
        return layout.interpret(tree, self.model)

    def iterdecode(self,
                   lines: Union[Iterable[str], str]) -> Iterator[Graph]:
        """
        Yield graphs parsed from *lines*.

        Args:
            lines: a string or open file with PENMAN-serialized graphs
        Returns:
            The :class:`~penman.graph.Graph` objects described in
            *lines*.
        """
        for tree in iterparse(lines):
            yield layout.interpret(tree, self.model)

    def iterparse(self, lines: Union[Iterable[str], str]) -> Iterator[Tree]:
        """
        Yield trees parsed from *lines*.

        Args:
            lines: a string or open file with PENMAN-serialized graphs
        Returns:
            The :class:`~penman.tree.Tree` object described in
            *lines*.
        """
        yield from iterparse(lines)

    def parse(self, s: str) -> Tree:
        """
        Parse PENMAN-notation string *s* into its tree structure.

        Args:
            s: a string containing a single PENMAN-serialized graph
        Returns:
            The tree structure described by *s*.
        Example:
            >>> from penman.codec import PENMANCodec
            >>> codec = PENMANCodec()
            >>> codec.parse('(b / bark-01 :ARG0 (d / dog))')  # noqa
            Tree(('b', [('/', 'bark-01'), (':ARG0', ('d', [('/', 'dog')]))]))
        """
        return parse(s)

    def parse_triples(self, s: str) -> List[BasicTriple]:
        """ Parse a triple conjunction from *s*."""
        return parse_triples(s)

    def encode(self,
               g: Graph,
               top: Variable = None,
               indent: Union[int, None] = -1,
               compact: bool = False) -> str:
        """
        Serialize the graph *g* into PENMAN notation.

        Args:
            g: the Graph object
            top: if given, the node to use as the top in serialization
            indent: how to indent formatted strings
            compact: if ``True``, put initial attributes on the first line
        Returns:
            the PENMAN-serialized string of the Graph *g*
        Example:
            >>> from penman.graph import Graph
            >>> from penman.codec import PENMANCodec
            >>> codec = PENMANCodec()
            >>> codec.encode(Graph([('h', 'instance', 'hi')]))
            '(h / hi)'

        """
        tree = layout.configure(g, top=top, model=self.model)
        return self.format(tree, indent=indent, compact=compact)

    def format(self,
               tree: Tree,
               indent: Union[int, None] = -1,
               compact: bool = False) -> str:
        """
        Format *tree* into a PENMAN string.
        """
        return format(tree, indent=indent, compact=compact)

    def format_triples(self,
                       triples: Iterable[BasicTriple],
                       indent: bool = True) -> str:
        """
        Return the formatted triple conjunction of *triples*.

        Args:
            triples: an iterable of triples
            indent: how to indent formatted strings
        Returns:
            the serialized triple conjunction of *triples*
        Example:
            >>> from penman.codec import PENMANCodec
            >>> codec = PENMANCodec()
            >>> codec.format_triples([('a', ':instance', 'alpha'),
            ...                       ('a', ':ARG0', 'b'),
            ...                       ('b', ':instance', 'beta')])
            ...
            'instance(a, alpha) ^\\nARG0(a, b) ^\\ninstance(b, beta)'

        """
        return format_triples(triples, indent=indent)


# The following are for the top-level API. They are renamed when they
# are imported into __init__.py. They are named with the leading
# underscore here so they are not included as part of penman.codec's
# public API.

def _decode(s: str,
            model: Model = None) -> Graph:
    """
    Deserialize PENMAN-serialized *s* into its Graph object

    Args:
        s: a string containing a single PENMAN-serialized graph
        model: the model used for interpreting the graph
    Returns:
        the Graph object described by *s*
    Example:
        >>> import penman
        >>> penman.decode('(b / bark-01 :ARG0 (d / dog))')
        <Graph object (top=b) at ...>

    """
    codec = PENMANCodec(model=model)
    return codec.decode(s)


def _iterdecode(lines: Union[Iterable[str], str],
                model: Model = None) -> Iterator[Graph]:
    """
    Yield graphs parsed from *lines*.

    Args:
        lines: a string or open file with PENMAN-serialized graphs
        model: the model used for interpreting the graph
    Returns:
        The :class:`~penman.graph.Graph` objects described in
        *lines*.
    Example:
        >>> import penman
        >>> for g in penman.iterdecode('(a / alpha) (b / beta)'):
        ...     print(repr(g))
        <Graph object (top=a) at ...>
        <Graph object (top=b) at ...>
    """
    codec = PENMANCodec(model=model)
    yield from codec.iterdecode(lines)


def _encode(g: Graph,
            top: Variable = None,
            model: Model = None,
            indent: Union[int, bool] = -1,
            compact: bool = False) -> str:
    """
    Serialize the graph *g* from *top* to PENMAN notation.

    Args:
        g: the Graph object
        top: if given, the node to use as the top in serialization
        model: the model used for interpreting the graph
        indent: how to indent formatted strings
        compact: if ``True``, put initial attributes on the first line
    Returns:
        the PENMAN-serialized string of the Graph *g*
    Example:
        >>> import penman
        >>> from penman.graph import Graph
        >>> penman.encode(Graph([('h', 'instance', 'hi')]))
        '(h / hi)'

    """
    codec = PENMANCodec(model=model)
    return codec.encode(g,
                        top=top,
                        indent=indent,
                        compact=compact)


def _load(source: FileOrFilename,
          model: Model = None) -> List[Graph]:
    """
    Deserialize a list of PENMAN-encoded graphs from *source*.

    Args:
        source: a filename or file-like object to read from
        model: the model used for interpreting the graph
    Returns:
        a list of Graph objects
    """
    codec = PENMANCodec(model=model)
    if isinstance(source, (str, Path)):
        with open(source) as fh:
            return list(codec.iterdecode(fh))
    else:
        assert hasattr(source, 'read')
        return list(codec.iterdecode(source))


def _loads(string: str,
           model: Model = None) -> List[Graph]:
    """
    Deserialize a list of PENMAN-encoded graphs from *string*.

    Args:
        string: a string containing graph data
        model: the model used for interpreting the graph
    Returns:
        a list of Graph objects
    """
    codec = PENMANCodec(model=model)
    return list(codec.iterdecode(string))


def _dump(graphs: Iterable[Graph],
          file: FileOrFilename,
          model: Model = None,
          indent: Union[int, bool] = -1,
          compact: bool = False) -> None:
    """
    Serialize each graph in *graphs* to PENMAN and write to *file*.

    Args:
        graphs: an iterable of Graph objects
        file: a filename or file-like object to write to
        model: the model used for interpreting the graph
        indent: how to indent formatted strings
        compact: if ``True``, put initial attributes on the first line
    """
    codec = PENMANCodec(model=model)
    if isinstance(file, (str, Path)):
        with open(file, 'w') as fh:
            _dump_stream(fh, graphs, codec, indent, compact)
    else:
        assert hasattr(file, 'write')
        _dump_stream(file, graphs, codec, indent, compact)


def _dump_stream(fh, gs, codec, indent, compact):
    """Helper method for dump() for incremental printing."""
    ss = (codec.encode(g, indent=indent, compact=compact)
          for g in gs)
    try:
        print(next(ss), file=fh)
    except StopIteration:
        return
    for s in ss:
        print(file=fh)
        print(s, file=fh)


def _dumps(graphs: Iterable[Graph],
           model: Model = None,
           indent: Union[int, bool] = -1,
           compact: bool = False) -> str:
    """
    Serialize each graph in *graphs* to the PENMAN format.

    Args:
        graphs: an iterable of Graph objects
        model: the model used for interpreting the graph
        indent: how to indent formatted strings
        compact: if ``True``, put initial attributes on the first line
    Returns:
        the string of serialized graphs
    """
    codec = PENMANCodec(model=model)
    strings = [codec.encode(g, indent=indent, compact=compact)
               for g in graphs]
    return '\n\n'.join(strings)



"""
No-op semantic model definition.
"""

from penman.types import BasicTriple
from penman.model import Model


class NoOpModel(Model):
    """
    A no-operation model that mostly leaves things alone.

    This model is like the default :class:`~penman.model.Model` except
    that :meth:`NoOpModel.deinvert` always returns the original
    triple, even if it was inverted.
    """
    def deinvert(self, triple: BasicTriple) -> BasicTriple:
        """Return *triple* (does not deinvert)."""
        return triple


model = NoOpModel()




"""
Functions for basic reading and writing of PENMAN graphs.

NOTE: This module is now deprecated and will be removed in a future
version. Its functions are now available in the penman module.
"""

import warnings

from penman.codec import (  # noqa: F401
    _decode as decode,
    _iterdecode as iterdecode,
    _encode as encode,
    _load as load,
    _loads as loads,
    _dump as dump,
    _dumps as dumps,
)

warnings.warn(
    'The penman.interface module is deprecated. Use the functions from '
    'the penman module directly, e.g., penman.decode().',
    DeprecationWarning)




from typing import Union, Iterable, Iterator, List
import logging

from penman.types import (
    Target,
    BasicTriple,
)
from penman.tree import Tree
from penman._lexer import (
    PENMAN_RE,
    TRIPLE_RE,
    lex,
    TokenIterator,
)


logger = logging.getLogger('pe')


def parse(s: str) -> Tree:
    """
    Parse PENMAN-notation string *s* into its tree structure.

    Args:
        s: a string containing a single PENMAN-serialized graph
    Returns:
        The tree structure described by *s*.
    Example:
        >>> import penman
        >>> penman.parse('(b / bark-01 :ARG0 (d / dog))')  # noqa
        Tree(('b', [('/', 'bark-01'), (':ARG0', ('d', [('/', 'dog')]))]))

    """
    tokens = lex(s, pattern=PENMAN_RE)
    return _parse(tokens)


def iterparse(lines: Union[Iterable[str], str]) -> Iterator[Tree]:
    """
    Yield trees parsed from *lines*.

    Args:
        lines: a string or open file with PENMAN-serialized graphs
    Returns:
        The :class:`~penman.tree.Tree` object described in *lines*.
    Example:
        >>> import penman
        >>> for t in penman.iterparse('(a / alpha) (b / beta)'):
        ...     print(repr(t))
        ...
        Tree(('a', [('/', 'alpha')]))
        Tree(('b', [('/', 'beta')]))

    """
    tokens = lex(lines, pattern=PENMAN_RE)
    while tokens and tokens.peek().type in ('COMMENT', 'LPAREN'):
        yield _parse(tokens)


def parse_triples(s: str) -> List[BasicTriple]:
    """
    Parse a triple conjunction from *s*.

    Example:
        >>> import penman
        >>> for triple in penman.parse_triples('''
        ...         instance(b, bark) ^
        ...         ARG0(b, d) ^
        ...         instance(d, dog)'''):
        ...     print(triple)
        ('b', ':instance', 'bark')
        ('b', ':ARG0', 'd')
        ('d', ':instance', 'dog')

        """
    tokens = lex(s, pattern=TRIPLE_RE)
    return _parse_triples(tokens)


def _parse(tokens: TokenIterator) -> Tree:
    metadata = _parse_comments(tokens)
    node = _parse_node(tokens)
    tree = Tree(node, metadata=metadata)
    logger.debug('Parsed: %s', tree)
    return tree


def _parse_comments(tokens: TokenIterator):
    """
    Parse PENMAN comments from *tokens* and return any metadata.
    """
    metadata = {}
    while tokens.peek().type == 'COMMENT':
        comment = tokens.next().text
        while comment:
            comment, found, meta = comment.rpartition('::')
            if found:
                key, _, value = meta.partition(' ')
                metadata[key] = value.rstrip()
    return metadata


def _parse_node(tokens: TokenIterator):
    """
    Parse a PENMAN node from *tokens*.

    Nodes have the following pattern::

        Node := '(' ID ('/' Concept)? Edge* ')'
    """
    tokens.expect('LPAREN')

    var = None
    concept: Union[str, None]
    edges = []

    if tokens.peek().type != 'RPAREN':
        var = tokens.expect('SYMBOL').text
        if tokens.peek().type == 'SLASH':
            slash = tokens.next()
            # for robustness, don't assume next token is the concept
            if tokens.peek().type in ('SYMBOL', 'STRING'):
                concept = tokens.next().text
                if tokens.peek().type == 'ALIGNMENT':
                    concept += tokens.next().text
            else:
                concept = None
                logger.warning('Missing concept: %s', slash.line)
            edges.append(('/', concept))
        while tokens.peek().type != 'RPAREN':
            edges.append(_parse_edge(tokens))

    tokens.expect('RPAREN')

    return (var, edges)


def _parse_edge(tokens: TokenIterator):
    """
    Parse a PENMAN edge from *tokens*.

    Edges have the following pattern::

        Edge := Role (Constant | Node)
    """
    role_token = tokens.expect('ROLE')
    role = role_token.text
    if tokens.peek().type == 'ALIGNMENT':
        role += tokens.next().text

    target = None
    _next = tokens.peek()
    next_type = _next.type
    if next_type in ('SYMBOL', 'STRING'):
        target = tokens.next().text
        if tokens.peek().type == 'ALIGNMENT':
            target += tokens.next().text
    elif next_type == 'LPAREN':
        target = _parse_node(tokens)
    # for robustness in parsing, allow edges with no target:
    #    (x :ROLE :ROLE2...  <- followed by another role
    #    (x :ROLE )          <- end of node
    elif next_type not in ('ROLE', 'RPAREN'):
        raise tokens.error('Expected: SYMBOL, STRING, LPAREN', token=_next)
    else:
        logger.warning('Missing target: %s', role_token.line)

    return (role, target)


def _parse_triples(tokens: TokenIterator) -> List[BasicTriple]:
    target: Target
    triples = []
    strip_caret = False
    while True:
        role = tokens.expect('SYMBOL').text
        if strip_caret and role.startswith('^'):
            role = role[1:]
        if not role.startswith(':'):
            role = ':' + role
        tokens.expect('LPAREN')
        symbol = tokens.expect('SYMBOL')
        source, target = _parse_triple(symbol, tokens)
        tokens.expect('RPAREN')

        if target is None:
            logger.warning('Triple without a target: %s', symbol.line)

        triples.append((source, role, target))

        # continue only if triple is followed by ^
        if tokens:
            _next = tokens.peek()
            if _next.type != 'SYMBOL' or not _next.text.startswith('^'):
                break
            elif _next.text == '^':
                strip_caret = False
                tokens.next()
            else:
                strip_caret = True
        else:
            break
    return triples


def _parse_triple(symbol, tokens):
    # SYMBOL may contain commas, so handle it here. If there
    # is no space between the comma and the next SYMBOL, they
    # will be grouped as one.
    source, comma, rest = symbol.text.partition(',')
    target = None
    if rest:  # role(a,b)
        target = rest
    else:
        if comma:  # role(a, b) OR role(a,)
            _next = tokens.accept('SYMBOL')
            if _next:
                target = _next.text
        else:  # role(a , b) OR role(a ,b) OR role(a ,) OR role(a)
            _next = tokens.accept('SYMBOL')
            if not _next:  # role(a)
                pass
            elif _next.text == ',':  # role(a , b) OR role(a ,)
                _next = tokens.accept('SYMBOL')
                if _next:  # role(a , b)
                    target = _next.text
            elif _next.text.startswith(','):  # role(a ,b)
                target = _next.text[1:]
            else:  # role(a b)
                tokens.error("Expected: ','", token=_next)
    return source, target



# -*- coding: utf-8 -*-

"""
Penman graph library.
"""

# the following is to retain parts of the original API

__all__ = [
    '__version__',
    '__version_info__',
    'PenmanError',
    'DecodeError',
    'Tree',
    'Triple',
    'Graph',
    'PENMANCodec',
    'parse',
    'iterparse',
    'parse_triples',
    'format',
    'format_triples',
    'interpret',
    'configure',
    'decode',
    'iterdecode',
    'encode',
    'load',
    'loads',
    'dump',
    'dumps',
]

import logging

from penman.__about__ import (
    __version__,
    __version_info__,
)
from penman.exceptions import (
    PenmanError,
    DecodeError,
)
from penman.tree import Tree
from penman.graph import (
    Triple,
    Graph,
)
from penman.layout import (
    interpret,
    configure,
)
from penman._parse import (
    parse,
    iterparse,
    parse_triples,
)
from penman._format import (
    format,
    format_triples,
)
from penman.codec import (
    PENMANCodec,
    _decode as decode,
    _iterdecode as iterdecode,
    _encode as encode,
    _load as load,
    _loads as loads,
    _dump as dump,
    _dumps as dumps,
)

logging.basicConfig()




from typing import Optional, Union, List, Iterable

from penman.types import BasicTriple
from penman.tree import (Tree, is_atomic)


def format(tree: Tree,
           indent: Union[int, None] = -1,
           compact: bool = False) -> str:
    """
    Format *tree* into a PENMAN string.

    Args:
        tree: a Tree object
        indent: how to indent formatted strings
        compact: if ``True``, put initial attributes on the first line
    Returns:
        the PENMAN-serialized string of the Tree *t*
    Example:
        >>> import penman
        >>> print(penman.format(
        ...     ('b', [('/', 'bark-01'),
        ...            (':ARG0', ('d', [('/', 'dog')]))])))
        (b / bark-01
           :ARG0 (d / dog))
    """
    if not isinstance(tree, Tree):
        tree = Tree(tree)
    vars = [var for var, _ in tree.nodes()] if compact else []
    parts = ['# ::{}{}'.format(key, ' ' + value if value else value)
             for key, value in tree.metadata.items()]
    parts.append(_format_node(tree.node, indent, 0, set(vars)))
    return '\n'.join(parts)


def format_triples(triples: Iterable[BasicTriple], indent: bool = True) -> str:
    """
    Return the formatted triple conjunction of *triples*.

    Args:
        triples: an iterable of triples
        indent: how to indent formatted strings
    Returns:
        the serialized triple conjunction of *triples*
    Example:
        >>> import penman
        >>> g = penman.decode('(b / bark-01 :ARG0 (d / dog))')
        >>> print(penman.format_triples(g.triples))
        instance(b, bark-01) ^
        ARG0(b, d) ^
        instance(d, dog)

    """
    delim = ' ^\n' if indent else ' ^ '
    # need to remove initial : on roles for triples
    conjunction = [f'{role.lstrip(":")}({source}, {target})'
                   for source, role, target in triples]
    return delim.join(conjunction)


def _format_node(node,
                 indent: Optional[int],
                 column: int,
                 vars: set) -> str:
    """
    Format tree *node* into a PENMAN string.
    """
    var, edges = node
    if not var:
        return '()'  # empty node
    if not edges:
        return f'({var!s})'  # var-only node

    # determine appropriate joiner based on value of indent
    if indent is None:
        joiner = ' '
    else:
        if indent == -1:
            column += len(str(var)) + 2  # +2 for ( and a space
        else:
            column += indent
        joiner = '\n' + ' ' * column

    # format the edges and join them
    # if vars is non-empty, all initial attributes are compactly
    # joined on the same line, otherwise they use joiner
    parts: List[str] = []
    compact = bool(vars)
    for edge in edges:
        target = edge[1]
        if compact and (not is_atomic(target) or target in vars):
            compact = False
            if parts:
                parts = [' '.join(parts)]
        parts.append(_format_edge(edge, indent, column, vars))
    # check if all edges can be compactly written
    if compact:
        parts = [' '.join(parts)]

    return f'({var!s} {joiner.join(parts)})'


def _format_edge(edge, indent, column, vars):
    """
    Format tree *edge* into a PENMAN string.
    """
    role, target = edge

    if role != '/' and not role.startswith(':'):
        role = ':' + role

    if indent == -1:
        column += len(role) + 1  # +1 for :

    sep = ' '
    if not target:
        target = sep = ''
    elif not is_atomic(target):
        target = _format_node(target, indent, column, vars)

    return f'{role}{sep}{target!s}'
